from abc import ABC, abstractmethod
from typing import Sequence, Union, Any, Optional
from queue import Queue

from smartpipeline.item import DataItem

__author__ = "Giacomo Berardi <giacbrd.com>"


class NameMixin:
    """
    Simple mixin for setting a name to an object
    """

    def set_name(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return getattr(self, "_name", f"{self.__class__.name}_{id(self)}")


class ConcurrentMixin:
    def on_fork(self) -> Any:
        """
        Called after concurrent stage executor initialization in a process (multiprocessing concurrency).
        The stage in the executor is a copy of the original,
        by overriding this method one can initialize variables specifically for the copies.
        """
        pass


class Processor(ABC):
    @abstractmethod
    def process(self, item: DataItem) -> DataItem:
        """
        Process a single item received by the stage.
        Must be overridden for properly defining a stage

        :return: The same item instance processed and enriched by the stage
        """
        return item


class BatchProcessor(ABC):
    @abstractmethod
    def process_batch(self, items: Sequence[DataItem]) -> Sequence[DataItem]:
        """
        Process a batch of items received by the stage.
        Must be overridden for properly defining a batch stage

        :return: The same batch with items processed and enriched by the stage
        """
        return items


class Stage(NameMixin, ConcurrentMixin, Processor):
    """
    Extend this class and override :meth:`.Stage.process` for defining a stage
    """

    def __str__(self) -> str:
        return "Stage {}".format(self.name)


class BatchStage(NameMixin, ConcurrentMixin, BatchProcessor):
    """
    Extend this class and override :meth:`.BatchStage.process_batch` for defining a batch stage
    """

    def __init__(self, size: int, timeout: Optional[float] = None):
        """
        :param size: Maximum size of item batches that can be processed together
        :param timeout: Seconds to wait before flushing a batch (calling :meth:`.BatchStage.process_batch` on it)
        """
        self._size = size
        self._timeout = timeout

    def __str__(self) -> str:
        return "Batch stage {}".format(self.name)

    @property
    def size(self) -> int:
        """
        Get the maximum size of item batches that can be processed together
        """
        return self._size

    @property
    def timeout(self) -> Optional[float]:
        """
        Seconds to wait before flushing a batch (calling :meth:`.BatchStage.process_batch` on it)
        """
        return self._timeout


class Source(ABC):
    """
    Extend this for defining a pipeline source
    """

    @abstractmethod
    def pop(self) -> Optional[DataItem]:
        """
        Generate items for feeding a pipeline.
        Must be overridden for properly defining a source.
        Call :meth:`.Source.stop` when item generation is ended

        :return: The generated item, if None it is simply ignored (e.g. after calling :meth:`.Source.stop`)
        """
        pass

    def get_item(self, block: bool = False) -> Optional[DataItem]:
        return self.pop()

    def stop(self):
        """
        Declare the end item generation, this event will be spread through the pipeline
        """
        self._is_stopped = True

    @property
    def is_stopped(self) -> bool:
        """
        True if the source has called the stop event
        """
        return getattr(self, "_is_stopped", False)


ItemsQueue = "Queue[DataItem]"
StageType = Union[Stage, BatchStage]



from __future__ import annotations
from typing import Optional


class Error(Exception):
    """
    Base exception type which usually only provokes skipping a stage for an item
    """

    def set_stage(self, stage: str):
        self._pipeline_stage = stage

    def get_stage(self) -> Optional[str]:
        return getattr(self, "_pipeline_stage", None)

    def with_exception(self, exception: Exception) -> Error:
        """
        Set the original exception (if any) that has generated this error
        """
        self._pipeline_exception = exception
        return self

    def get_exception(self) -> Exception:
        """
        Get the original exception (if any) that has generated this error
        """
        return getattr(self, "_pipeline_exception", Exception())

    def __str__(self) -> str:
        error = super(Error, self).__str__().strip()
        exception = str(self.get_exception()).strip()
        return "\n".join((error, exception)).strip()


class CriticalError(Error):
    """
    A type of exception which usually provokes skipping the whole pipeline for an item
    """

    pass



import threading
from abc import ABC, abstractmethod
from collections import OrderedDict

__author__ = "Giacomo Berardi <giacbrd.com>"


class LastOrderedDict(OrderedDict):
    def last_key(self):
        return next(reversed(self.keys()))


class ConcurrentCounter(ABC):
    """
    Interface for a counter that is safe for concurrent access
    """

    @abstractmethod
    def __iadd__(self, incr):
        return self

    @abstractmethod
    def value(self):
        return 0


class ThreadCounter(ConcurrentCounter):
    """
    Thread safe counter
    """

    def __init__(self):
        self._value = 0
        self._lock = threading.Lock()

    def __iadd__(self, incr):
        with self._lock:
            self._value += incr
        return self

    @property
    def value(self):
        with self._lock:
            return self._value


class ProcessCounter(ConcurrentCounter):
    """
    Process safe counter
    """

    def __init__(self, manager):
        self._value = manager.Value("i", 0)
        self._lock = manager.Lock()

    def __iadd__(self, incr):
        with self._lock:
            self._value.value += incr
        return self

    @property
    def value(self):
        with self._lock:
            return self._value.value



from __future__ import annotations
import logging
from typing import Optional

from smartpipeline.error.exceptions import Error, CriticalError
from smartpipeline.item import DataItem
from smartpipeline.stage import NameMixin

__author__ = "Giacomo Berardi <giacbrd.com>"


class ErrorManager:
    """
    Basic error handling of a pipeline, principally manages :class:`.exceptions.Error` and :class:`.exceptions.CriticalError` types
    """

    def __init__(self):
        self._raise_on_critical = False
        self._skip_on_critical = True
        self._logger = logging.getLogger(self.__class__.__name__)

    def raise_on_critical_error(self) -> ErrorManager:
        """
        If a :class:`.exceptions.CriticalError` or any un-managed exception is met, raise it externally, kill the pipeline
        """
        self._raise_on_critical = True
        return self

    def no_skip_on_critical_error(self) -> ErrorManager:
        """
        Change default behaviour on :class:`.exceptions.CriticalError`: just skip the current stage .
        Valid only if :meth:`.ErrorManager.raise_on_critical_error` is not set
        """
        self._skip_on_critical = False
        return self

    def handle(
        self, error: Exception, stage: NameMixin, item: DataItem
    ) -> Optional[CriticalError]:
        """
        Manage an error produced by a stage

        :param error: It can be a generic exception or an error from :mod:`.exceptions` explicitly raised by a stage
        :param stage: Stage which raised the exception during processing
        :param item: Item which raised the exception when processed
        :return: If the handled error results to be critical return the generated :class:`.exceptions.CriticalError`
        """
        if type(error) is Error:
            item_error = item.add_error(stage.name, error)
        else:
            # any un-managed exception is a potential critical error
            item_error = item.add_critical_error(stage.name, error)
        exc_info = (type(error), error, error.__traceback__)
        self._logger.exception(self._generate_message(stage, item), exc_info=exc_info)
        if isinstance(item_error, CriticalError):
            exception = self._check_critical(item_error)
            if exception:
                return item_error

    @staticmethod
    def _generate_message(stage: NameMixin, item: DataItem) -> str:
        return "The stage {} has generated an error on item {}".format(stage, item)

    def _check_critical(self, error: CriticalError) -> Optional[Exception]:
        """
        Manage a critical error, usually after an item has been processed by a stage

        :return: The exception which caused a critical error if any
        """
        ex = error.get_exception()
        if self._raise_on_critical:
            raise ex
        elif self._skip_on_critical:
            return ex

    def check_critical_errors(self, item: DataItem) -> Optional[Exception]:
        """
        Check the critical errors registered for an item
        """
        if item.has_critical_errors():
            for er in item.critical_errors():
                ex = self._check_critical(er)
                if ex:
                    return ex



from __future__ import annotations

from typing import Union, Generator, Any, KeysView, Callable, Dict
from smartpipeline.defaults import PAYLOAD_SNIPPET_SIZE
from smartpipeline.error.exceptions import Error, CriticalError

__author__ = "Giacomo Berardi <giacbrd.com>"


class DataItem:
    """
    An item containing a unit of data to process through a pipeline.
    Is ia usually generated by the :class:`.stage.Source` subclass and retrieved at the end of a :class:`.pipeline.Pipeline`.
    """

    def __init__(self):
        self._errors = []
        self._critical_errors = []
        self._meta = {}
        self._payload = {}
        self._timings = {}
        self._callback_fun = None

    def __str__(self) -> str:
        return f"Data item {self.id} with payload {self.payload_snippet()}..."

    @property
    def payload(self) -> Dict[str, Any]:
        """
        Access to the actual data contained in the item

        :return: A dictionary in which organizing data by fields (recommended to be JSON serializable)
        """
        return self._payload

    def payload_snippet(self, max_size: int = PAYLOAD_SNIPPET_SIZE):
        """
        A short string representation of the :attr:`.DataItem.payload` (recommended to override this method)

        :param max_size: Maximum size of the string representation
        """
        return str(self.payload)[:max_size]

    def set_metadata(self, field: str, value: Any) -> DataItem:
        """
        Add a metadata, something we want to remember but keeping it outside the :attr:`.DataItem.payload`

        :param field: Name of the metadata variable
        """
        self._meta[field] = value
        return self

    def get_metadata(self, field: str) -> Any:
        """
        Get a metadata value by its name

        :return: A value or None if the metadata does not exist in the item
        """
        return self._meta.get(field)

    @property
    def metadata_fields(self) -> KeysView[str]:
        """
        Get all the metadata names defined in this item
        """
        return self._meta.keys()

    def set_timing(self, stage_name: str, milliseconds: float) -> DataItem:
        """
        Set the time spent by a stage (referenced by its name) for processing the item
        """
        self._timings[stage_name] = milliseconds
        return self

    def get_timing(self, stage_name: str) -> float:
        """
        Get the time spent by a stage (referenced by its name) for processing the item

        :return: The time in milliseconds or None if the item has not ben processed by the stage
        """
        return self._timings.get(stage_name)

    @property
    def timed_stages(self) -> KeysView[str]:
        """
        Get the stage names for which the item has a process time
        """
        return self._timings.keys()

    @property
    def id(self) -> Any:
        """
        Get the unique identifier of the item.
        It is recommended to override this in order to properly compute it from the :attr:`.DataItem.payload`
        """
        ret = self._payload.get("id")
        if ret is None:
            ret = self._meta.get("id")
            if ret is None:
                ret = id(self)
        return ret

    def set_callback(self, fun: Callable[[DataItem], Any]):
        """
        Set a function to call after a successful asynchronous execution of a pipeline on the item (through :meth:`.pipeline.Pipeline.process_async`)
        """
        self._callback_fun = fun

    def callback(self) -> Any:
        """
        Call the function set with :meth:`.DataItem.set_callback`
        """
        if self._callback_fun is not None:
            return self._callback_fun(self)

    def has_errors(self) -> bool:
        """
        True if the item has raised an :class:`.error.exceptions.Error` in some stage processing
        """
        return any(self._errors)

    def has_critical_errors(self) -> bool:
        """
        True if the item has raised a :class:`.error.exceptions.CriticalError` or any un-managed exception in some stage processing
        """
        return any(self._critical_errors)

    def errors(self) -> Generator[Error, None, None]:
        """
        Iter over :class:`.error.exceptions.Error` instances eventually generated in some stage processing
        """
        for e in self._errors:
            yield e

    def critical_errors(self) -> Generator[CriticalError, None, None]:
        """
        Iter over :class:`.error.exceptions.CriticalError` instances or any un-managed exception eventually generated in some stage processing
        """
        for e in self._critical_errors:
            yield e

    def add_error(self, stage: str, exception: Union[Error, Exception]) -> Error:
        """
        Add an :class:`.error.exceptions.Error` generated in a stage (referenced by its name) for the item

        :param exception: It can be an :class:`.error.exceptions.Error` instance or any exception, which will be encapsulated in an :class:`.error.exceptions.Error`
        """
        # if it is an instance of `Error`
        if type(exception) is Error:
            exception.set_stage(stage)
            self._errors.append(exception)
            return exception
        elif isinstance(exception, Exception) and type(exception) is not CriticalError:
            error = Error()
            error.with_exception(exception)
            error.set_stage(stage)
            self._errors.append(error)
            return error
        else:
            raise ValueError("Add a pipeline Error or a generic exception")

    def add_critical_error(
        self, stage: str, exception: Union[CriticalError, Exception]
    ) -> CriticalError:
        """
        Add a :class:`.error.exceptions.CriticalError` generated in a stage (referenced by its name) for the item

        :param exception: It can be a :class:`.error.exceptions.CriticalError` instance or any exception, which will be encapsulated in a :class:`.error.exceptions.CriticalError`
        """
        # if it is an instance of `CriticalError`
        if type(exception) is CriticalError:
            exception.set_stage(stage)
            self._critical_errors.append(exception)
            return exception
        elif isinstance(exception, Exception) and type(exception) is not Error:
            error = CriticalError()
            error.with_exception(exception)
            error.set_stage(stage)
            self._critical_errors.append(error)
            return error
        else:
            raise ValueError("Add a pipeline CriticalError or a generic exception")


class Stop(DataItem):
    """
    Core "signal" item (do not use this explicitly) used in pipelines for determining the end of a flow of items,
    like an event passed through all the stages
    """

    def __str__(self) -> str:
        return "Stop signal {}".format(self.id)



import os
from typing import Generator, Optional, Any, IO

from smartpipeline.stage import Source
from smartpipeline.item import DataItem

__author__ = "Giacomo Berardi <giacbrd.com>"


class LocalFilesSource(Source):
    """
    Generate a special file item for any file found in a directory
    """

    def __init__(self, dir_path: str, postfix: str = ""):
        self.postfix = postfix
        self.dir_path = dir_path
        self._iterator = self._iter_files()

    def _iter_files(self) -> Generator[str, None, None]:
        for fname in os.listdir(self.dir_path):
            if fname.endswith(self.postfix) and not fname.startswith("."):
                yield os.path.join(self.dir_path, fname)

    def pop(self) -> Optional[DataItem]:
        file_path = next(self._iterator, None)
        if file_path:
            item = FilePathItem(file_path)
            return item
        else:
            self.stop()


class FilePathItem(DataItem):
    """
    An item with a pointer to a file path
    """

    def __init__(self, path: str):
        super().__init__()
        self.path = path

    def __str__(self) -> str:
        return f"Data item {self.id}, file path {self.path}, with payload {self.payload_snippet()}..."

    @property
    def id(self) -> Any:
        return os.path.basename(self.path) or super().id()


class FileObjItem(FilePathItem):
    """
    An item with a pointer to a file object
    """

    def __init__(self, file: IO):
        super().__init__(file.name)
        self.file = file



__author__ = "Giacomo Berardi <giacbrd.com>"

CONCURRENCY_WAIT: float = 0.1  # seconds to wait in loops querying queues or threads
PAYLOAD_SNIPPET_SIZE: int = 100  # size of of the string defining a snippet of an item payload
MAX_QUEUES_SIZE: int = 1000  # default size for queues between concurrent stages, let's avoid infinite queues



"""
Containers encapsulate stages and manage their execution
"""

import concurrent
import queue
from queue import Queue
from abc import ABC, abstractmethod
from concurrent.futures import wait, Executor, Future
from concurrent.futures.process import ProcessPoolExecutor
from concurrent.futures.thread import ThreadPoolExecutor
from threading import Event
from typing import Sequence, Optional, Callable
from smartpipeline.utils import ConcurrentCounter
from smartpipeline.error.handling import ErrorManager
from smartpipeline.executors import (
    process,
    process_batch,
    stage_executor,
    batch_stage_executor,
    StageExecutor,
)
from smartpipeline.stage import Stage, BatchStage, Source, ItemsQueue, StageType
from smartpipeline.item import DataItem, Stop

__author__ = "Giacomo Berardi <giacbrd.com>"


QueueInitializer = Callable[[], ItemsQueue]
CounterInitializer = Callable[[], ConcurrentCounter]
EventInitializer = Callable[[], Event]


class InQueued(ABC):
    """
    Interface for containers which exposes an output queue for items
    """

    @property
    @abstractmethod
    def out_queue(self) -> ItemsQueue:
        """
        Get the output queue instance
        """
        pass

    @abstractmethod
    def init_queue(self, initializer: QueueInitializer):
        """
        Initialize the output queue with a specific constructor (function)
        """
        pass


class BaseContainer(InQueued):
    """
    Base interface for all containers
    """

    def __init__(self):
        self._is_stopped = False
        self._is_terminated = False
        self._out_queue = None
        self._counter = 0

    @abstractmethod
    def get_processed(
        self, block: bool = False, timeout: Optional[int] = None
    ) -> Optional[DataItem]:
        """
        Get the oldest processed item waiting to be retrieved

        :param block: Wait for the next item to be processed if no one available
        :param timeout: Time to wait when `block` is True
        :return: A processed item or None if: no item waiting to be retrieved; timeout expires on a blocked call
        """
        pass

    def init_queue(self, initializer: QueueInitializer) -> ItemsQueue:
        self._out_queue = initializer()
        return self._out_queue

    @property
    def out_queue(self) -> ItemsQueue:
        return self._out_queue

    def is_stopped(self) -> bool:
        return self._is_stopped

    def stop(self):
        """
        Set the container stage as stopped, usually when the pipeline has ended its work (i.e. the source has called :meth:`.stage.Source.stop`)
        """
        self._is_stopped = True

    def is_terminated(self) -> bool:
        return self._is_terminated

    def terminate(self):
        """
        Turn off the container (e.g. threads and processes) and the relative stage
        """
        self._is_terminated = True

    def count(self) -> int:
        """
        Return the number of items that have been seen by this container
        """
        return self._counter

    def increase_count(self):
        """
        Increase the counter of items "seen" by the container/stage
        """
        self._counter += 1


class ConnectedStageMixin:
    """
    A mixin for the containers that encapsulate stages that can be connected to others (i.e. in a pipeline)
    """

    @property
    def previous(self) -> BaseContainer:
        """
        Get the container from which this container receives items
        """
        return self._previous

    def set_previous(self, container: BaseContainer):
        """
        Set the container from which this container will receive items
        """
        self._previous = container


class FallibleMixin:
    """
    A mixin for the containers that encapsulate stages that can produce errors during processing
    """

    def set_error_manager(self, error_manager: ErrorManager):
        self._error_manager = error_manager

    @property
    def error_manager(self) -> ErrorManager:
        return self._error_manager


class NamedStageMixin:
    """
    A mixin for basic containers of stages
    """

    def set_stage(self, name: str, stage: StageType):
        self._name = name
        self._stage = stage
        self._stage.set_name(name)

    @property
    def name(self) -> str:
        return self._name

    @property
    def stage(self) -> StageType:
        return self._stage


class SourceContainer(BaseContainer):
    """
    A container specific for sources
    """

    def __init__(self):
        super().__init__()
        self._source = None
        # the next two attributes are used jointly for "manually" prepending items
        self._next_item = None
        self._internal_queue_obj = None
        self._stop_sent = False
        self._queue_initializer = None

    @property
    def _internal_queue(self) -> ItemsQueue:
        """
        A special queue used for "manually" enrich the source with items
        """
        if self._internal_queue_obj is None:
            if self._queue_initializer is None:
                self._set_internal_queue_initializer()
            self._internal_queue_obj = self._queue_initializer()
        return self._internal_queue_obj

    def _set_internal_queue_initializer(self, initializer: ItemsQueue = queue.Queue):
        self._queue_initializer = initializer

    def __str__(self) -> str:
        return "Base container for source {}".format(self._source)

    def set(self, source: Source):
        """
        Set the actual source for this container
        """
        self._source = source

    def is_set(self) -> bool:
        """
        True if this container is ready to produce items in output
        """
        return (
            self._source is not None
            or self._next_item is not None
            or self._out_queue is not None
        )

    def is_stopped(self) -> bool:
        """
        True if this container has ended the production of items
        """
        if self._source is not None:
            return getattr(self._source, "is_stopped", False)
        else:
            return self._is_stopped

    def stop(self):
        """
        Stop this source, the container won't produce items anymore
        """
        if self._source is not None:
            self._source.stop()
        self._is_stopped = True

    def pop_into_queue(self):
        """
        Pop from the source but put the item in the queue that will be read from the first stage of the pipeline.
        Only used with concurrent stages
        """
        while True:
            item = self._get_next_item()
            if self._stop_sent:
                return
            elif item is None:
                continue
            else:
                self.out_queue.put(item, block=True)
                if not isinstance(item, Stop):
                    self.increase_count()
            if isinstance(item, Stop):
                self._stop_sent = True
                return

    def prepend_item(self, item: Optional[DataItem]):
        """
        Enrich the source with items "manually".
        Only used for processing single items
        """
        if self._next_item is not None:
            self._internal_queue.put(item)
        else:
            self._next_item = item

    def get_processed(
        self, block: bool = True, timeout: Optional[int] = None
    ) -> Optional[DataItem]:
        if self.out_queue is not None:
            item = self.out_queue.get(block=block, timeout=timeout)
        else:
            item = self._get_next_item()
            if item is not None and not isinstance(item, Stop):
                self.increase_count()
        return item

    def _get_next_item(self) -> DataItem:
        """
        Obtain the next item to send to output according to the source status and "manually" added items
        """
        ret = self._next_item
        self._next_item = None
        if ret is not None:
            try:
                self._next_item = self._internal_queue.get(block=False)
                self._internal_queue.task_done()
            except queue.Empty:
                if self.is_stopped():
                    self._next_item = Stop()
            return ret
        elif self._source is not None:
            ret = self._source.pop()
            if self.is_stopped():
                return Stop()
            else:
                return ret
        elif self.is_stopped():
            return Stop()


class StageContainer(
    BaseContainer, NamedStageMixin, FallibleMixin, ConnectedStageMixin
):
    """
    The standard container for basic stages
    """

    def __init__(self, name: str, stage: Stage, error_manager: ErrorManager):
        super().__init__()
        self.set_error_manager(error_manager)
        self.set_stage(name, stage)
        self._last_processed = None

    def __str__(self) -> str:
        return "Container for stage {}".format(self._stage)

    def process(self) -> DataItem:
        item = self.previous.get_processed()
        if isinstance(item, Stop):
            self.stop()
        elif item is not None:
            item = process(self.stage, item, self._error_manager)
        self._put_item(item)
        return item

    def get_processed(
        self, block: bool = False, timeout: Optional[int] = None
    ) -> Optional[DataItem]:
        ret = self._last_processed
        self._last_processed = None
        # if we are in a concurrent stage the items are obtained exclusively from the output queue
        if self.out_queue is not None:
            try:
                ret = self.out_queue.get(block=block, timeout=timeout)
                self.out_queue.task_done()
            except queue.Empty:
                return None
        return ret

    def _put_item(self, item: DataItem):
        """
        A processed item is set as next output.
        If we are processing asynchronously (e.g. concurrent stage) it is put in the output queue,
        otherwise a reference to the this last processed item is set
        """
        self._last_processed = item
        if (
            self.out_queue is not None
            and self._last_processed is not None
            and not self.is_terminated()
        ):
            self.out_queue.put(self._last_processed, block=True)
        if item is not None and not isinstance(item, Stop):
            self.increase_count()


class BatchStageContainer(
    BaseContainer, NamedStageMixin, FallibleMixin, ConnectedStageMixin
):
    """
    Container for batch stages
    """

    def __init__(self, name: str, stage: BatchStage, error_manager: ErrorManager):
        super().__init__()
        self.set_error_manager(error_manager)
        self.set_stage(name, stage)
        # TODO next two varibales are for non-concurrent container, that is currently never used, it doesn't work
        self.__result_queue: ItemsQueue = queue.SimpleQueue()
        self._last_processed: Sequence[DataItem] = []

    def process(self) -> Sequence[DataItem]:
        items = []
        # items that we want to put as last in a batch, ergo in output
        extra_items = []
        for _ in range(self.stage.size):
            item = self.previous.get_processed(timeout=self.stage.timeout)
            if isinstance(item, Stop):
                self.stop()
                self._put_item([item])
                extra_items.append(item)
            elif item is not None:
                items.append(item)
        if any(items):
            items = process_batch(self.stage, items, self.error_manager)
            self._put_item(items)
        return items + extra_items

    def get_processed(
        self, block: bool = False, timeout: Optional[int] = None
    ) -> Optional[DataItem]:
        if (
            self.out_queue is not None
            and self.__result_queue.qsize() < self.out_queue.qsize()
        ):
            try:
                # let's free the output queue (so this stage con continue processing) and keep the items internally
                for _ in range(self.stage.size):
                    item = self.out_queue.get(block=block, timeout=timeout)
                    if item is not None:
                        self.__result_queue.put_nowait(item)
                    self.out_queue.task_done()
            except queue.Empty:
                try:
                    return self.__result_queue.get(block=block, timeout=timeout)
                except queue.Empty:
                    return None
        elif self._last_processed:
            for item in self._last_processed:
                self.__result_queue.put_nowait(item)
            self._last_processed = []
        try:
            return self.__result_queue.get(block=block, timeout=timeout)
        except queue.Empty:
            return None

    def _put_item(self, items: Sequence[DataItem]):
        """
        A batch of processed items is set as next output.
        If we are processing asynchronously (e.g. concurrent stage) they are put in the output queue,
        otherwise a list of last processed items is extended
        """
        self._last_processed.extend(items)
        if (
            self.out_queue is not None
            and self._last_processed is not None
            and not self.is_terminated()
        ):
            for item in self._last_processed:
                self.out_queue.put(item, block=True)
        for item in self._last_processed:
            if item is not None and not isinstance(item, Stop):
                self.increase_count()

    @property
    def size(self) -> int:
        return self.stage.size


class ConcurrentContainer(InQueued, ConnectedStageMixin):
    """
    Base container for stages that must process concurrently and asynchronously
    """

    def init_concurrency(
        self,
        queue_initializer: QueueInitializer,
        counter_initializer: CounterInitializer,
        terminate_event_initializer: EventInitializer,
        concurrency: int = 1,
        use_threads: bool = True,
    ):
        """
        Initialization of instance members

        :param queue_initializer: Constructor for output, and eventually input, queue
        :param counter_initializer: Constructor for items counter, it counts items seen by concurrent stage executions
        :param terminate_event_initializer: Constructor for the event for alerting all concurrent stage executions for termination
        :param concurrency: Number of maximum concurrent stage executions
        :param use_threads: True for using threads for concurrency, otherwise use multiprocess
        """
        self._concurrency = concurrency
        self._use_threads = use_threads
        self._stage_executor = None
        self._previous_queue = None
        self._futures: Sequence[Future] = []
        self._queue_initializer = queue_initializer
        self._out_queue = self._queue_initializer()
        self._counter = counter_initializer()
        self._counter_initializer = counter_initializer
        self._terminate_event = terminate_event_initializer()

    def queues_empty(self) -> bool:
        """
        True of both input and output queues are empty
        """
        return self._previous_queue.empty() and self.out_queue.empty()

    def queues_join(self):
        self._previous_queue.join()
        self.out_queue.join()

    def empty_queues(self):
        """
        Delete all items in both input and output queues
        """
        for q in (self._previous_queue, self.out_queue):
            while True:
                try:
                    q.get_nowait()
                    q.task_done()
                except queue.Empty:
                    break

    def set_previous(self, container: BaseContainer):
        self._previous = container
        if (
            isinstance(
                self._previous,
                (ConcurrentStageContainer, BatchConcurrentStageContainer),
            )
            and not self._previous.use_threads
        ):
            # give priority to the previous queue initializer
            self._previous_queue = self._previous.out_queue
        else:
            self._previous_queue = self._previous.init_queue(self._queue_initializer)

    def _get_stage_executor(self) -> Executor:
        """
        Get and eventually generate a pool executor where concurrent stage executions run
        """
        if self._stage_executor is None:
            executor = ThreadPoolExecutor if self._use_threads else ProcessPoolExecutor
            # TODO one executor per stage? why max_workers are equivalent to concurrency?
            self._stage_executor = executor(max_workers=self._concurrency)
        return self._stage_executor

    @property
    def use_threads(self) -> bool:
        """
        True if we are using threads, False if we are using multiprocess
        """
        return self._use_threads

    def shutdown(self):
        """
        Turn off the container pool executor
        """
        for future in self._futures:
            future.cancel()
        if self._stage_executor is not None:
            self._stage_executor.shutdown()

    def __del__(self):
        self.shutdown()

    def check_errors(self):
        """
        Look for exceptions generated by stage executions inside threads/processes, eventually re-raise exceptions
        """
        for future in self._futures:
            try:
                ex = future.exception(timeout=0)
            except concurrent.futures.TimeoutError:
                continue
            if ex is not None:
                raise ex

    def count(self) -> int:
        return self._counter.value if self._counter else 0

    def terminate(self):
        """
        Alert stage executors for termination
        """
        self._terminate_event.set()
        wait(self._futures)

    def is_terminated(self) -> bool:
        """
        Check if termination has been set and all stage executions have been terminated
        """
        return (
            all(future.done() or future.cancelled() for future in self._futures)
            and self._terminate_event.is_set()
        )

    def _run(
        self,
        stage: StageType,
        _executor: StageExecutor,
        in_queue: ItemsQueue,
        out_queue: ItemsQueue,
        error_manager: ErrorManager,
    ):
        """
        Start the concurrent execution of stage processing.
        The stage will consume and produce from input/output queues concurrently

        :param stage: Stage instance
        :param _executor: Function to run concurrently in the executor, which performs the stage executions
        :param in_queue: Previous stage output queue
        :param out_queue: Output queue
        :param error_manager: Error manager instance
        """
        ex = self._get_stage_executor()
        self._counter = self._counter_initializer()
        self._terminate_event.clear()
        for _ in range(self._concurrency):
            self._futures.append(
                ex.submit(
                    _executor,
                    stage,
                    in_queue,
                    out_queue,
                    error_manager,
                    self._terminate_event,
                    self._counter,
                )
            )


class ConcurrentStageContainer(ConcurrentContainer, StageContainer):
    """
    Standard stage container with concurrency
    """

    def __init__(
        self,
        name: str,
        stage: Stage,
        error_manager: ErrorManager,
        queue_initializer: QueueInitializer,
        counter_initializer: CounterInitializer,
        terminate_event_initializer: EventInitializer,
        concurrency: int = 1,
        use_threads: bool = True,
    ):
        """
        :param name: Stage name
        :param stage: Stage instance
        :param error_manager: Error manager instance
        :param queue_initializer: Constructor for output, and eventually input, queue
        :param counter_initializer: Constructor for items counter, it counts items seen by concurrent stage executions
        :param terminate_event_initializer: Constructor for the event for alerting all concurrent stage executions for termination
        :param concurrency: Number of maximum concurrent stage executions
        :param use_threads: True for using threads for concurrency, otherwise use multiprocess
        """
        super().__init__(name, stage, error_manager)
        self.init_concurrency(
            queue_initializer,
            counter_initializer,
            terminate_event_initializer,
            concurrency,
            use_threads,
        )

    def run(self, _executor: StageExecutor = stage_executor):
        """
        Start the concurrent execution of stage processing.
        The stage will consume and produce from input/output queues concurrently

        :param _executor: Function to run in the executor, which performs the stage executions
        """
        super()._run(
            self.stage,
            _executor,
            self._previous_queue,
            self.out_queue,
            self.error_manager,
        )


class BatchConcurrentStageContainer(ConcurrentContainer, BatchStageContainer):
    """
    Batch stage container with concurrency
    """

    def __init__(
        self,
        name: str,
        stage: BatchStage,
        error_manager: ErrorManager,
        queue_initializer: QueueInitializer,
        counter_initializer: CounterInitializer,
        terminate_event_initializer: EventInitializer,
        concurrency: int = 1,
        use_threads: bool = True,
    ):
        """
        :param name: Stage name
        :param stage: Stage instance
        :param error_manager: Error manager instance
        :param queue_initializer: Constructor for output, and eventually input, queue
        :param counter_initializer: Constructor for items counter, it counts items seen by concurrent stage executions
        :param terminate_event_initializer: Constructor for the event for alerting all concurrent stage executions for termination
        :param concurrency: Number of maximum concurrent stage executions
        :param use_threads: True for using threads for concurrency, otherwise use multiprocess
        """
        super().__init__(name, stage, error_manager)
        self.init_concurrency(
            queue_initializer,
            counter_initializer,
            terminate_event_initializer,
            concurrency,
            use_threads,
        )

    def run(self, _executor: StageExecutor = batch_stage_executor):
        """
        Start the concurrent execution of stage processing.
        The stage will consume and produce from input/output queues concurrently

        :param _executor: Function to run in the executor, which performs the stage executions
        """
        super()._run(
            self.stage,
            _executor,
            self._previous_queue,
            self.out_queue,
            self.error_manager,
        )



import queue
import time
from threading import Event
from typing import Sequence, Callable, Optional, List
from smartpipeline.utils import ConcurrentCounter, ProcessCounter
from smartpipeline.defaults import CONCURRENCY_WAIT
from smartpipeline.error.handling import ErrorManager
from smartpipeline.item import DataItem, Stop
from smartpipeline.stage import Stage, BatchStage, ItemsQueue, StageType

__author__ = "Giacomo Berardi <giacbrd.com>"


def process(stage: Stage, item: DataItem, error_manager: ErrorManager) -> DataItem:
    """
    Execute the :meth:`.stage.Stage.process` method of a stage for an item
    """
    if error_manager.check_critical_errors(item):
        return item
    time1 = time.time()
    try:
        ret = stage.process(item)
    except Exception as e:
        item.set_timing(stage.name, (time.time() - time1) * 1000.0)
        error_manager.handle(e, stage, item)
        return item
    # this can't be in a finally, otherwise it would register the `error_manager.handle` time
    item.set_timing(stage.name, (time.time() - time1) * 1000.0)
    return ret


def process_batch(
    stage: BatchStage, items: Sequence[DataItem], error_manager: ErrorManager
) -> List[Optional[DataItem]]:
    """
    Execute the :meth:`.stage.BatchStage.process_batch` method of a batch stage for a batch of items
    """
    ret: List[Optional[DataItem]] = [None] * len(items)
    to_process = {}
    for i, item in enumerate(items):
        if error_manager.check_critical_errors(item):
            ret[i] = item
        else:
            to_process[i] = item
    time1 = time.time()
    try:
        processed = stage.process_batch(list(to_process.values()))
    except Exception as e:
        spent = ((time.time() - time1) * 1000.0) / (len(to_process) or 1.0)
        for i, item in to_process.items():
            item.set_timing(stage.name, spent)
            error_manager.handle(e, stage, item)
            ret[i] = item
        return ret
    spent = ((time.time() - time1) * 1000.0) / (len(to_process) or 1.0)
    for n, i in enumerate(to_process.keys()):
        item = processed[n]
        item.set_timing(stage.name, spent)
        ret[i] = item
    return ret


def stage_executor(
    stage: Stage,
    in_queue: ItemsQueue,
    out_queue: ItemsQueue,
    error_manager: ErrorManager,
    terminated: Event,
    counter: ConcurrentCounter,
):
    """
    Consume items from an input queue, process and put them in a output queue, indefinitely,
    until a termination event is set
    """
    if isinstance(counter, ProcessCounter):
        # call this only if the stage is a copy of the original, ergo it is executed in a process
        stage.on_fork()
    while True:
        if terminated.is_set() and in_queue.empty():
            return
        try:
            item = in_queue.get(block=True, timeout=CONCURRENCY_WAIT)
        except queue.Empty:
            continue
        if isinstance(item, Stop):
            out_queue.put(item, block=True)
            in_queue.task_done()
        elif item is not None:
            try:
                item = process(stage, item, error_manager)
            except Exception as e:
                raise e
            else:
                if item is not None:
                    out_queue.put(item, block=True)
                    if not isinstance(item, Stop):
                        counter += 1
            finally:
                in_queue.task_done()


def batch_stage_executor(
    stage: BatchStage,
    in_queue: ItemsQueue,
    out_queue: ItemsQueue,
    error_manager: ErrorManager,
    terminated: Event,
    counter: ConcurrentCounter,
):
    """
    Consume items in batches from an input queue, process and put them in a output queue, indefinitely,
    until a termination event is set
    """
    if isinstance(counter, ProcessCounter):
        # call this only if the stage is a copy of the original, ergo it is executed in a process
        stage.on_fork()
    while True:
        if terminated.is_set() and in_queue.empty():
            return
        items = []
        try:
            for _ in range(stage.size):
                item = in_queue.get(block=True, timeout=stage.timeout)
                # give priority to the Stop event item
                if isinstance(item, Stop):
                    out_queue.put(item, block=True)
                elif item is not None:
                    items.append(item)
                in_queue.task_done()
        except queue.Empty:
            if not any(items):
                continue
        if any(items):
            try:
                items = process_batch(stage, items, error_manager)
            except Exception as e:
                raise e
            else:
                for item in items:
                    if item is not None:
                        out_queue.put(item, block=True)
                        if not isinstance(item, Stop):
                            counter += 1


StageExecutor = Callable[
    [StageType, ItemsQueue, ItemsQueue, ErrorManager, Event, ConcurrentCounter], None
]



from __future__ import annotations

import time
from concurrent.futures._base import Future, Executor
from concurrent.futures.process import ProcessPoolExecutor
from concurrent.futures.thread import ThreadPoolExecutor
from multiprocessing import Manager
from queue import Queue
from threading import Thread, Event
from typing import Generator, Optional, Callable, Sequence, Mapping, Any

from smartpipeline.defaults import CONCURRENCY_WAIT, MAX_QUEUES_SIZE
from smartpipeline.error.handling import ErrorManager
from smartpipeline.containers import (
    SourceContainer,
    StageContainer,
    ConcurrentStageContainer,
    BatchStageContainer,
    BatchConcurrentStageContainer,
    BaseContainer,
    ConnectedStageMixin,
)
from smartpipeline.stage import BatchStage, ItemsQueue, Source, StageType
from smartpipeline.item import Stop, DataItem
from smartpipeline.utils import LastOrderedDict, ThreadCounter, ProcessCounter

__author__ = "Giacomo Berardi <giacbrd.com>"


class FakeContainer:
    def __init__(self, item: DataItem):
        self._item = item

    def get_item(self) -> DataItem:
        return self._item


class Pipeline:
    def __init__(
        self,
        max_init_workers: Optional[int] = None,
        max_queues_size: int = MAX_QUEUES_SIZE,
    ):
        """
        :param max_init_workers: Number of workers to use for concurrent initialization of stages, default the number of CPUs
        :param max_queues_size: Maximum size of any queue instanced for the pipeline (stage input and output queues)
        """
        self._containers = LastOrderedDict()
        self._error_manager = ErrorManager()
        self._max_init_workers = max_init_workers
        self._init_executor = None
        self._wait_previous_executor = None
        self._pipeline_executor = None
        self._max_queues_size = max_queues_size
        self._out_queue = None
        self._enqueue_source = False
        self._sync_manager = None
        # an empty source, on which we can only occasionally send items
        self._source_container = SourceContainer()
        self._count = 0
        self._executors_ready = False

    def _new_mp_queue(self) -> ItemsQueue:
        """
        Construct queue for multiprocessing communication
        """
        if self._sync_manager is None:
            self._sync_manager = Manager()
        return self._sync_manager.Queue(maxsize=self._max_queues_size)

    def _new_queue(self) -> ItemsQueue:
        """
        Construct queue for communication
        """
        return Queue(maxsize=self._max_queues_size)

    def _new_mp_event(self) -> Event:
        """
        Construct synchronization event for multiprocessing
        """
        if self._sync_manager is None:
            self._sync_manager = Manager()
        return self._sync_manager.Event()

    @staticmethod
    def _new_event() -> Event:
        """
        Construct synchronization event
        """
        return Event()

    def _new_mp_counter(self) -> ProcessCounter:
        """
        Construct a safe counter for multiprocessing
        """
        if self._sync_manager is None:
            self._sync_manager = Manager()
        return ProcessCounter(self._sync_manager)

    @staticmethod
    def _new_counter() -> ThreadCounter:
        """
        Construct a safe counter for threads
        """
        return ThreadCounter()

    def _wait_executors(self, wait_seconds: float = CONCURRENCY_WAIT):
        """
        Wait for all containers to start

        :param wait_seconds: Recurrently wait these seconds for all stage initializers to finish
        """
        if self._executors_ready:
            return
        if self._init_executor is not None:
            self._init_executor.shutdown(wait=True)
            self._init_executor = None
        while not all(self._containers.values()):
            time.sleep(wait_seconds)
        self._wait_previous_executor.shutdown(wait=True)
        for name, stage in self._containers.items():
            if isinstance(
                stage, (ConcurrentStageContainer, BatchConcurrentStageContainer)
            ):
                stage.run()
        self._executors_ready = True

    def shutdown(self):
        if self._out_queue is not None:
            self._out_queue.join()
        # if self._init_executor is not None:
        #     self._init_executor.shutdown()
        # FIXME stage shutdown may raise exception, the executor gets stuck
        # for name, stage in self._containers.items():
        #     if isinstance(stage, (ConcurrentStageContainer, BatchConcurrentStageContainer)):
        #         stage.shutdown()
        if self._sync_manager is not None:
            self._sync_manager.shutdown()

    def __del__(self):
        self.shutdown()

    def run(self) -> Generator[DataItem, None, None]:
        """
        Run the pipeline given a source and a concatenation of stages.
        Get the sequence of items through iteration

        :return: Iterator over processed items
        :raises ValueError: When a source has not been set for the pipeline
        """
        counter = 0
        self._wait_executors()
        if not self._source_container.is_set():
            raise ValueError("Set the data source for this pipeline")
        last_stage_name = self._last_stage_name()
        terminator_thread = None
        source_thread = None
        # in case the first stage is concurrent
        if self._enqueue_source:
            source_thread = Thread(target=self._source_container.pop_into_queue)
            source_thread.start()
        while True:
            for name, container in self._containers.items():
                try:
                    # concurrent stages run by themselves in threads/processes
                    if not isinstance(
                        container,
                        (ConcurrentStageContainer, BatchConcurrentStageContainer),
                    ):
                        container.process()
                    else:
                        container.check_errors()
                except Exception as e:
                    self.stop()
                    # TODO in case of errors we loose pending items!
                    self._terminate_all(force=True)
                    self.shutdown()
                    self._count += 1
                    raise e
                # retrieve finally processed items from the last stage
                if name == last_stage_name:
                    for _ in range(
                        container.size
                        if isinstance(container, BatchStageContainer)
                        else 1
                    ):
                        item = container.get_processed()
                        if item is not None:
                            if not isinstance(item, Stop):
                                yield item
                                counter += 1
                                self._count += 1
                            # if a stop is finally signaled, start termination of all containers
                            elif (
                                not self._all_terminated() and terminator_thread is None
                            ):
                                terminator_thread = Thread(target=self._terminate_all)
                                terminator_thread.start()
                        # an item is None if the final output queue is empty
                        else:
                            break
            # exit the loop only when all items have been returned
            if self._all_empty() and counter >= self._source_container.count():
                if source_thread is not None:
                    source_thread.join()
                if terminator_thread is not None:
                    terminator_thread.join()
                    self.shutdown()
                return

    @property
    def count(self) -> int:
        """
        Get the number of items processed by all executed runs, also for items which have failed

        :return: Count of processed items
        """
        return self._count

    def _terminate_all(
        self, force: bool = False, wait_seconds: float = CONCURRENCY_WAIT
    ):
        """
        Terminate all running containers

        :param force: If True do not wait for a container to process all items produced by the source
        :param wait_seconds: Time to wait before pinging again a container for its termination
        """
        # scroll the pipeline by its order and terminate stages after the relative queues are empty
        for container in self._containers.values():
            if not force:
                # ensure the stage has processed all source items
                while container.count() < self._source_container.count():
                    time.sleep(wait_seconds)
            container.terminate()
            if isinstance(container, ConcurrentStageContainer):
                if force:
                    # empty the queues, losing pending items
                    container.empty_queues()
                while not container.is_terminated():
                    time.sleep(wait_seconds)
                container.queues_join()
                while not container.queues_empty():
                    time.sleep(wait_seconds)

    def _all_terminated(self) -> bool:
        """
        Check if all containers have been alerted for termination and are exited
        """
        return all(container.is_terminated() for container in self._containers.values())

    def _all_empty(self) -> bool:
        """
        Check if all containers are terminated and there are not items left in the queues
        """
        return self._all_terminated() and all(
            container.queues_empty()
            for container in self._containers.values()
            if isinstance(
                container, (ConcurrentStageContainer, BatchConcurrentStageContainer)
            )
        )

    def process(self, item: DataItem) -> DataItem:
        """
        Process a single item synchronously (no concurrency) through the pipeline
        """
        self._wait_executors()
        last_stage_name = self._containers.last_key()
        self._source_container.prepend_item(item)
        for name, container in self._containers.items():
            container.process()
            if name == last_stage_name:
                return container.get_processed(block=True)

    def process_async(
        self, item: DataItem, callback: Optional[Callable[[DataItem], Any]] = None
    ):
        """
        Process a single item asynchronously through the pipeline, stages may run concurrently.
        The call returns immediately, processed items are retrieved with :meth:`.Pipeline.get_item`

        :param callback: A function to call after a successful process of the item
        """
        self._wait_executors()
        if callback is not None:
            item.set_callback(callback)
        self._source_container.prepend_item(item)
        self._start_pipeline_executor()

    def stop(self):
        """
        Tell the source to stop to generate items and consequently the pipeline
        """
        self._source_container.stop()

    def get_item(self, block: bool = True) -> DataItem:
        """
        Get a single item from the asynchronous execution of the pipeline on single items from :meth:`.Pipeline.process_async`

        :param block: If True wait indefinitely for the next processed item
        :raises ValueError: When there is not output queue set, the pipeline is not running asynchronously
        :raises queue.Empty: When we do not block and the queue is empty
        """
        if self._out_queue is not None:
            item = self._out_queue.get(block)
            self._out_queue.task_done()
            return item
        else:
            raise ValueError(
                "No pipeline is running asynchronously, not item can be retrieved from the output queue"
            )

    def set_source(self, source: Source) -> Pipeline:
        """
        Set the source of the pipeline: a subclass of :class:`.stage.Source`
        """
        self._source_container.set(source)
        return self

    def set_error_manager(self, error_manager: ErrorManager) -> Pipeline:
        """
        Set the error manager for handling errors from each stage item processing
        """
        self._error_manager = error_manager
        for container in self._containers.values():
            container.set_error_manager(self._error_manager)
        return self

    def _last_stage_name(self) -> str:
        if self._containers:
            return self._containers.last_key()

    def _last_container(self) -> BaseContainer:
        if self._containers:
            return self._containers[self._last_stage_name()]
        else:
            return self._source_container

    def _wait_for_previous(
        self,
        container: ConnectedStageMixin,
        last_stage_name: str,
        wait_seconds: float = CONCURRENCY_WAIT,
    ):
        """
        Given a container we want to append to the pipeline, wait for the last one (added to the pipeline) to be created

        :param container: A container to add to the pipeline
        :param last_stage_name: Name of the last stage currently in the pipeline
        :param wait_seconds: Time to recurrently wait the construction of the container relative to the last stage in the pipeline
        """

        def _waiter():
            if last_stage_name is not None:
                while self._containers[last_stage_name] is None:
                    time.sleep(wait_seconds)
                container.set_previous(self._containers[last_stage_name])
            else:
                container.set_previous(self._source_container)

        executor = self._get_wait_previous_executor()
        executor.submit(_waiter)

    def _get_container(
        self, name: str, stage: StageType, concurrency: int, use_threads: bool
    ) -> BaseContainer:
        """
        Get a new container instance according to the pipeline configuration

        :param name: Stage name
        :param stage: A stage instance
        :param concurrency: Number of concurrent stage executions, if 0 then just create the non-concurrent containers
        :param use_threads: If True use threads, otherwise multiprocessing
        """
        if concurrency <= 0:
            constructor = (
                BatchStageContainer if isinstance(stage, BatchStage) else StageContainer
            )
            return constructor(name, stage, self._error_manager)
        else:
            constructor = (
                BatchConcurrentStageContainer
                if isinstance(stage, BatchStage)
                else ConcurrentStageContainer
            )
            if use_threads:
                return constructor(
                    name,
                    stage,
                    self._error_manager,
                    self._new_queue,
                    self._new_counter,
                    self._new_event,
                    concurrency,
                    use_threads,
                )
            else:
                return constructor(
                    name,
                    stage,
                    self._error_manager,
                    self._new_mp_queue,
                    self._new_mp_counter,
                    self._new_mp_event,
                    concurrency,
                    use_threads,
                )

    def get_stage(self, name: str) -> StageType:
        """
        Get a stage instance by its name
        """
        return self._containers.get(name).stage

    def append_stage(
        self,
        name: str,
        stage: StageType,
        concurrency: int = 0,
        use_threads: bool = True,
    ) -> Pipeline:
        """
        Append a stage to the pipeline just after the last one appended, or after the source if it is the first stage

        :param name: Name for identify the stage in the pipeline, it is also set in the stage and it must be unique in the pipeline
        :param stage: Instance of a stage
        :param concurrency: Number of concurrent stage executions, if 0 then threads/processes won't be involved for this stage
        :param use_threads: If True use threads, otherwise multiprocessing
        """
        self._executors_ready = False
        # FIXME here we force a BatchStage to run on a thread, but we would leave it on the main thread
        if concurrency < 1 and isinstance(stage, BatchStage):
            use_threads = True
            concurrency = 1
        self._check_stage_name(name)
        container = self._get_container(name, stage, concurrency, use_threads)
        if concurrency > 0:
            # if it is concurrent and it is the first stage, make the source working on a output queue
            if not self._containers:
                self._enqueue_source = True
        self._wait_for_previous(
            container, self._last_stage_name()
        )  # wait that previous stage is initialized
        self._containers[name] = container
        return self

    def append_stage_concurrently(
        self,
        name: str,
        stage_class: Callable,
        args: Sequence = None,
        kwargs: Mapping = None,
        concurrency: int = 0,
        use_threads: bool = True,
    ) -> Pipeline:
        """
        Append a stage class to the pipeline just after the last one appended, or after the source if it is the first stage.
        The stage construction will be executed concurrently respect to the general pipeline construction

        :param name: Name for identify the stage in the pipeline, it is also set in the stage and it must be unique in the pipeline
        :param stage_class: Class of a stage
        :param args: List of arguments for the stage constructor
        :param kwargs: Dictionary of keyed arguments for the stage constructor
        :param concurrency: Number of concurrent stage executions, if 0 then threads/processes won't be involved for this stage
        :param use_threads: If True use threads, otherwise multiprocessing
        """
        self._executors_ready = False
        # FIXME here we force a BatchStage to run on a thread, but we would leave it on the main thread
        if concurrency < 1 and issubclass(stage_class, BatchStage):
            use_threads = True
            concurrency = 1
        if kwargs is None:
            kwargs = {}
        if args is None:
            args = []
        self._check_stage_name(name)
        # if it is concurrent and it is the first stage, make the source working on a output queue
        if concurrency > 0 and not self._containers:
            self._enqueue_source = True
        last_stage_name = self._last_stage_name()
        # set it immediately so the order of the calls of this method is followed in `_containers`
        self._containers[name] = None
        future = self._get_init_executor(use_threads).submit(
            stage_class, *args, **kwargs
        )

        def append_stage(stage_future: Future):
            stage = stage_future.result()
            container = self._get_container(name, stage, concurrency, use_threads)
            self._wait_for_previous(container, last_stage_name)
            self._containers[name] = container

        future.add_done_callback(append_stage)
        return self

    def _get_init_executor(self, use_threads: bool = True) -> Executor:
        """
        Get a pool executor for concurrent stage initialization

        :param use_threads: True if the executor uses treads, otherwise multiprocessing
        """
        if self._init_executor is None:
            executor = ThreadPoolExecutor if use_threads else ProcessPoolExecutor
            self._init_executor = executor(max_workers=self._max_init_workers)
        return self._init_executor

    def _get_wait_previous_executor(self) -> Executor:
        """
        Get a pool executor for the function that will recurrently wait for a container to be ready
        """
        if self._wait_previous_executor is None:
            self._wait_previous_executor = ThreadPoolExecutor()
        return self._wait_previous_executor

    def _start_pipeline_executor(self) -> Thread:
        """
        Get a thread where to run a pipeline that accepts asynchronous processing of single items
        """
        if self._pipeline_executor is None:
            self._init_out_queue()

            def pipeline_runner():
                for item in self.run():
                    item.callback()
                    self._out_queue.put(item)

            self._pipeline_executor = Thread(target=pipeline_runner, daemon=True)
            self._pipeline_executor.start()
        return self._pipeline_executor

    def _check_stage_name(self, name: str):
        """
        Check if a stage name is not already defined in the pipeline
        :raises ValueError: Stage name is already defined in the pipeline
        """
        if name in self._containers:
            raise ValueError(f"The stage name {name} is already used in this pipeline")

    def _init_out_queue(self):
        """
        Get the internal output pipeline for asynchronous processing of single items
        """
        self._out_queue = self._new_queue()



import os

from setuptools import setup, find_packages

__author__ = "Giacomo Berardi <giacbrd.com>"


def readfile(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    return open(path).read()


setup(
    name="SmartPipeline",
    version="0.3.5",
    description="A framework for fast developing scalable data pipelines following a simple design pattern",
    long_description=readfile("README.rst"),
    long_description_content_type="text/x-rst",
    classifiers=[
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 4 - Beta",
    ],
    url="https://github.com/giacbrd/SmartPipeline",
    project_urls={
        "Documentation": "https://smartpipeline.readthedocs.io",
        "Source": "https://github.com/giacbrd/SmartPipeline",
    },
    keywords=[
        "data pipeline",
        "task queue",
        "data science",
        "machine learning",
        "design pattern",
    ],
    author="Giacomo Berardi",
    author_email="barnets@gmail.com",
    packages=find_packages(exclude=["tests"]),
    tests_require=["pytest"],
    python_requires=">=3.7",
)
